# Playlist Manager.pm by Philip Meyer (phil@hergest.demon.co.uk)
# Version 1.0 - 14 January, 2005
# Version 1.1 - 29 May, 2006
#
# This code is derived from code with the following copyright message:
#
# SlimServer Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2.

use strict;

package Plugins::PlaylistMan;

use Slim::Utils::Strings qw (string);
use Slim::Utils::Misc;
use Slim::Player::Client;
use Slim::Player::Playlist;
use File::Spec::Functions qw(:ALL);

# Export the version to the server
use vars qw($VERSION);
#$VERSION = substr(q$Revision: 2.0 $,10);
$VERSION = "2.0";

use constant PLAYLISTMAN_NOW_PLAYING_MENU => 'PLUGIN_PLAYLISTMAN.NowPlayingAddMenu';
use constant PLAYLISTMAN_MENU_ADD => 'PLUGIN_PLAYLISTMAN.SelectPlaylistMenu';
use constant PLAYLISTMAN_SELECT_OTHER_PLAYLIST => 'PLUGIN_PLAYLISTMAN.SelectOtherPlaylist';

sub getDisplayName()
{
	return substr ($::VERSION, 0, 1) >= 6 ? 'PLUGIN_PLAYLISTMAN_NAME' : string('PLUGIN_PLAYLISTMAN_NAME');
}

sub strings()
{
    return "
PLUGIN_PLAYLISTMAN_NAME
	EN	Playlist Manager

PLUGIN_PLAYLISTMAN_SELECT_ADDOPTION
	EN	Playlist Manager

PLUGIN_PLAYLISTMAN_SELECT_ADDTO_OPTION
	EN	Add track to which playlist?

PLUGIN_PLAYLISTMAN_SAVE_NEW_PLAYLIST
	EN	Save to new playlist:

PLUGIN_PLAYLISTMAN_SAVE_NEW_PLAYLIST_CONFIRM
	EN	Confirm new playlist:
";
}

our @LegalChars = (
	Slim::Display::Display::symbol('rightarrow'),
	'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
	'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
	'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
	' ',
	'.', '-', '_',
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9'
);

our @legalMixed = (
	[' ','0'], 				# 0
	['.','-','_','1'], 			# 1
	['a','b','c','A','B','C','2'],		# 2
	['d','e','f','D','E','F','3'], 		# 3
	['g','h','i','G','H','I','4'], 		# 4
	['j','k','l','J','K','L','5'], 		# 5
	['m','n','o','M','N','O','6'], 		# 6
	['p','q','r','s','P','Q','R','S','7'], 	# 7
	['t','u','v','T','U','V','8'], 		# 8
	['w','x','y','z','W','X','Y','Z','9'] 	# 9
);


our %context = ();

my %positions = ();

my $playlistindex;
my $song;

sub enabled
{
	return 1;
}

sub setMode
{
	my $client = shift;
	my $method = shift || '';

	# Handle request to exit our mode.
	if ($method eq 'pop')
	{
		# Pop the current mode off the mode stack and restore the previous one
		Slim::Buttons::Common::popMode($client);
		return;
	}

	$positions {$client} = 0 if (!defined $positions{$client});

	$client->lines(\&lines);
}

sub lines
{
	my $client = shift;
	my ($line1, $line2);

	my @addoptions = ("Display Bad Tags", "Display Bad Music");

	$line1 = "Display Playlists";
	$line2 = $addoptions[$positions{$client}];

	return ($line1, $line2);
}

our %functions =
(
	'left' => sub  {
		my $client = shift;
		Slim::Buttons::Common::popModeRight($client);
	},
	'right' => sub {
		my $client = shift;
		$client->bumpRight();
	},
	'up' => sub {
		my $client = shift;
		my $newPos = Slim::Buttons::Common::scroll($client, -1, 2, $positions{$client});
		$positions{$client} = $newPos;
		$client->update();
	},
	'down' => sub {
		my $client = shift;
		my $newPos = Slim::Buttons::Common::scroll($client, +1, 2, $positions{$client});
		$positions{$client} = $newPos;
		$client->update();
	},
	'addmenu' => sub
	{
		my $client = shift;
		$playlistindex = Slim::Buttons::Playlist::browseplaylistindex($client);
		$song = Slim::Player::Playlist::song($client, $playlistindex);
		my $songName = $song->title;

		$::d_plugins && Slim::Utils::Misc::msg("PlaylistMan: display addmenu for current playlist track number $playlistindex - $songName\n");

		my %params =
		(
			# The header (first line) to display whilst in this mode.
			header => '{PLUGIN_PLAYLISTMAN_SELECT_ADDOPTION} {count}',

			# A reference to the list of items to display.
			listRef => ['Add track to ...', 'Zap track', 'Save Playlist ...'],

			# A unique name for this mode that won't actually get displayed anywhere.
			modeName => PLAYLISTMAN_NOW_PLAYING_MENU,

			parentMode => Slim::Buttons::Common::mode($client),

			onPlay => sub {
				my ($client, $item) = @_;
				NowPlayingAddAction($client, $item);
			},

			# An anonymous function that is called every time the user presses the RIGHT button.
			onRight => sub {
				my ($client, $item) = @_;
				NowPlayingAddAction($client, $item);
			},

			# These are all menu items and so have a right-arrow overlay
			overlayRef => sub {
				return [ undef, Slim::Display::Display::symbol('rightarrow') ];
			}
		);

		Slim::Buttons::Common::pushModeLeft($client, 'INPUT.Choice', \%params);
	}
);

sub NowPlayingAddAction
{
	my ($client, $item) = @_;

	$::d_plugins && Slim::Utils::Misc::msg("Now Playing Add Action: $item\n");
	
	if (Slim::Player::Playlist::count($client) > 0)
	{
		if ($item eq 'Add track to ...')
		{
			my %params =
			(
				# The header (first line) to display whilst in this mode.
				header => '{PLUGIN_PLAYLISTMAN_SELECT_ADDTO_OPTION} {count}',

				# A reference to the list of items to display.
				listRef => ['Bad Music', 'Bad Tags', 'Morning Alarm', 'Other ...', 'New Playlist'],

				# A unique name for this mode that won't actually get displayed anywhere.
				modeName => PLAYLISTMAN_MENU_ADD,

				parentMode => Slim::Buttons::Common::mode($client),

				onPlay => sub {
					my ($client, $item) = @_;
					AddTrackToAction($client, $item);
				},

				# An anonymous function that is called every time the user presses the RIGHT button.
				onRight => sub {
					my ($client, $item) = @_;
					AddTrackToAction($client, $item);
				},

				# These are all menu items and so have a right-arrow overlay
				overlayRef => sub {
					return [ undef, Slim::Display::Display::symbol('rightarrow') ];
				}
			);

			Slim::Buttons::Common::pushModeLeft($client, 'INPUT.Choice', \%params);
		}
		elsif ($item eq "Save Playlist ...")
		{
			if (!Slim::Utils::Prefs::get('playlistdir'))
			{
			}
			else
			{
				$context{$client} = $client->currentPlaylist ? Slim::Music::Info::standardTitle($client, $client->currentPlaylist) : 'A';
				Slim::Buttons::Common::pushModeLeft($client, 'SavePlaylist');
			}
		}
		elsif ($item eq "Zap track")
		{
			$client->showBriefly($client->string('ZAPPING_FROM_PLAYLIST'), Slim::Music::Info::standardTitle($client, $song), undef, 1);

			$client->execute(["playlist", "zap", $playlistindex]);
	 		Slim::Buttons::Common::popMode($client);
		}
	}
}

sub AddTrackToAction
{
	my ($client, $item) = @_;

	if ($item eq 'Other ...')
	{
		SelectPlaylistToAddTo($client);
	}
	elsif ($item eq 'New Playlist')
	{
		Slim::Buttons::Common::pushModeLeft($client, 'NewPlaylist');
	}
	else
	{
		AddTrackTo($client, $item);
	}
}

sub SelectPlaylistToAddTo
{
	my $client = shift;

	$::d_plugins && Slim::Utils::Misc::msg("Select Playlist to add track to\n");

	my @dirItems=();

	my $playlists = Slim::Schema->rs('Playlist')->browse;
	my $playlistsCount = 0;
	while (my $playlistObj = $playlists->next)
	{
		$::d_plugins && Slim::Utils::Misc::msg("PlaylistMan: Found playlistname=$playlistObj\n");

		#TODO: Remove configured common playlists from the list of selectable playlists?
		push @dirItems, $playlistObj->title;

		$playlistsCount = $playlistsCount+1;
	}
	
	$::d_plugins && Slim::Utils::Misc::msg("Number of playlists: $playlistsCount\n");
		
	if ($playlistsCount > 0)
	{
		my %params =
		(
			# The header (first line) to display whilst in this mode.
			# Handle large Font setting
			header => '{PLUGIN_PLAYLISTMAN_SELECT_ADDTO_OPTION} {count}',
			listRef => \@dirItems,
			stringHeader => 1,

			# A unique name for this mode that won't actually get displayed anywhere.
			modeName => PLAYLISTMAN_SELECT_OTHER_PLAYLIST,
			parentMode => Slim::Buttons::Common::mode($client),
			onPlay => sub {
					my ($client, $item) = @_;
					AddTrackTo($client, $item);
				},

			# An anonymous function that is called every time the user presses the RIGHT button.
			onRight => sub {
					my ($client, $item) = @_;
					AddTrackTo($client, $item);
				},

			# These are all menu items and so have a right-arrow overlay
			overlayRef => 
			#	sub {
			#		my ($client, $item) = @_;	
			#		$item = Slim::Music::Info::standardTitle($client,$item);
			#		return [undef, $item.Slim::Display::Display::symbol('rightarrow') ];
			#	}
				sub {
					return [ undef, Slim::Display::Display::symbol('rightarrow') ];
				}		);

		Slim::Buttons::Common::pushModeLeft($client, 'INPUT.Choice', \%params);
	}
	else
	{
		$client->showBriefly("No Playlists");
	}
}

sub AddTrackTo
{
	my ($client, $item) = @_;

	$::d_plugins && Slim::Utils::Misc::msg("Add Track To action: $item\n");
	
	if (Slim::Player::Playlist::count($client) > 0)
	{
		$client->showBriefly("Adding to $item playlist", Slim::Music::Info::standardTitle($client, $song), undef, 1);

		AddToPlaylist($client, $item, $song);

#TODO: popMode doesn't seem to work any more
		Slim::Buttons::Common::popMode($client);
		Slim::Buttons::Common::popMode($client);
	}
}

sub AddToPlaylist
{
	my ($client, $playlist, $song) = @_;

	$::d_plugins && Slim::Utils::Misc::msg("PlaylistMan: Playlist name is: $playlist\n");

	my $playlistTitle = Slim::Music::Info::plainTitle($playlist, undef);

	$::d_plugins && Slim::Utils::Misc::msg("PlaylistMan: Playlist plaintitle is: '$playlistTitle'\n");

	my $fileurl = catfile(Slim::Utils::Prefs::get('playlistdir'), $playlist . '.m3u');
	$fileurl = Slim::Utils::Misc::fileURLFromPath($fileurl);

	$::d_plugins && Slim::Utils::Misc::msg("PlaylistMan: fileurl=$fileurl\n");

	my $playlistObj = Slim::Schema->rs('Playlist')->updateOrCreate({
		'url' => $fileurl,
		'playlist' => $playlist,
		'attributes' => {
			'TITLE' => $playlistTitle,
			'CT'    => 'ssp',
		},
	});

	$::d_plugins && Slim::Utils::Misc::msg("PlaylistMan: PlaylistObj=$playlistObj\n");

# 	my @list = $playlistObj->tracks();
# 	push @list, $song;
# 
#  	$::d_plugins && Slim::Utils::Misc::msg("PlaylistMan: Playlist contains @list\n");
# 
# 	$playlistObj->setTracks(\@list);
# 	$playlistObj->update();

	Slim::Schema->toggleDebug(1);
	$playlistObj->appendTracks([ $song ]);
	Slim::Schema->toggleDebug(0);

	Slim::Player::Playlist::scheduleWriteOfPlaylist($client, $playlistObj);
}

sub getFunctions
{
	return \%functions;
}

#Save Playlist functionality

sub SetSavePlaylistMode
{
	my ($client, $push) = @_;

	$client->lines(\&SavePlaylistLines);
	if (!Slim::Utils::Prefs::get('playlistdir'))
	{
	}
	elsif ($push ne 'push')
	{
		my $playlist = '';
	}
	else
	{
		$context{$client} = $client->currentPlaylist ? Slim::Music::Info::standardTitle($client, $client->currentPlaylist) : 'A';
	
		Slim::Buttons::Common::pushMode($client,'INPUT.Text', {
			'callback' => \&Plugins::PlaylistMan::SavePluginCallback,
			'valueRef' => \$context{$client},
			'charsRef' => \@LegalChars,
			'numberLetterRef' => \@legalMixed,
			'header' => $client->string('PLAYLIST_AS'),
			'cursorPos' => 0
		});
	}
}

sub SetNewPlaylistMode
{
	my ($client, $push) = @_;

	$client->lines(\&SavePlaylistLines);
	if (!Slim::Utils::Prefs::get('playlistdir'))
	{
	}
	elsif ($push ne 'push')
	{
		my $playlist = '';
	}
	else
	{
		$context{$client} = $client->currentPlaylist ? Slim::Music::Info::standardTitle($client, $client->currentPlaylist) : 'A';

		Slim::Buttons::Common::pushMode($client,'INPUT.Text', {
			'callback' => \&Plugins::PlaylistMan::SavePluginCallback,
			'valueRef' => \$context{$client},
			'charsRef' => \@LegalChars,
			'numberLetterRef' => \@legalMixed,
			'header' => $client->string('PLUGIN_PLAYLISTMAN_SAVE_NEW_PLAYLIST'),
			'cursorPos' => 0
		});
	}
}

sub SavePlaylistLines
{
	my $client = shift;

	my ($line1, $line2, $arrow);

	if (!Slim::Utils::Prefs::get('playlistdir'))
	{
		$line1 = $client->string('NO_PLAYLIST_DIR');
		$line2 = $client->string('NO_PLAYLIST_DIR_MORE');
	}
	else
	{
		$line1 = $client->string('PLUGIN_PLAYLISTMAN_SAVE_NEW_PLAYLIST_CONFIRM');
		$line2 = $context{$client};
		$arrow = $client->symbols('rightarrow');
	}

	return
	{
		'line1'    => $line1,
		'line2'    => $line2, 
		'overlay2' => $arrow,
	};
}

our %SavePlaylistFunctions = (
	'left' => sub  {
		my $client = shift;
		Slim::Buttons::Common::popModeRight($client);
	},
	'right' => sub  {
		my $client = shift;
		my $playlistfile = $context{$client};
		Slim::Buttons::Common::setMode($client, 'playlist');
		SavePlaylist($client, $playlistfile);
	},
	'save' => sub {
		my $client = shift;
		Slim::Buttons::Common::pushModeLeft($client, 'SavePlaylist');
	},
);

our %NewPlaylistFunctions = (
	'left' => sub  {
		my $client = shift;
		Slim::Buttons::Common::popModeRight($client);
	},
	'right' => sub  {
		my $client = shift;
		my $playlistfile = $context{$client};
		Slim::Buttons::Common::setMode($client, 'playlist');
		NewPlaylist($client, $playlistfile);
	},
	'save' => sub {
		my $client = shift;
		Slim::Buttons::Common::pushModeLeft($client, 'NewPlaylist');
	},
);

sub GetSavePlaylistFunctions
{
	return \%SavePlaylistFunctions;
}

sub GetNewPlaylistFunctions
{
	return \%NewPlaylistFunctions;
}

sub NewPlaylist
{
	my ($client, $playlist) = @_;

	$::d_plugins && Slim::Utils::Misc::msg("Save to playlist $playlist\n");

	AddToPlaylist($client, $playlist, $song);

	$client->showBriefly( {
		'line1' => $client->string('PLAYLIST_SAVING'),
		'line2' => $playlist
	});
}

sub SavePlaylist
{
	my $client = shift;
	my $playlistfile = shift;

	$client->execute(['playlist', 'save', $playlistfile]);
	$client->showBriefly( {
		'line1' => $client->string('PLAYLIST_SAVING'),
		'line2' => $playlistfile
	});
}

my $rightarrow = Slim::Display::Display::symbol('rightarrow');

sub SavePluginCallback
{
	my ($client,$type) = @_;

 	if ($type eq 'nextChar')
 	{
 		$context{$client} =~ s/$rightarrow//;
 		Slim::Buttons::Common::popMode($client);
 		$client->pushLeft();
 	}
 	elsif ($type eq 'backspace')
 	{
 		Slim::Buttons::Common::popModeRight($client);
 		Slim::Buttons::Common::popModeRight($client);
 	}
 	else
 	{
 		$client->bumpRight();
 	}
}


####################################################################
# Adds a mapping for 'addmenu' function in Now Playing mode.
####################################################################
our %mapping = ('add.hold' => 'addmenu');

sub defaultMap
{
	return \%mapping; 
}

sub initPlugin
{
	$::d_plugins && Slim::Utils::Misc::msg("PlaylistMan: Initialise\n");

	Slim::Buttons::Common::addMode('SavePlaylist', GetSavePlaylistFunctions(), \&Plugins::PlaylistMan::SetSavePlaylistMode);
	Slim::Buttons::Common::addMode('NewPlaylist', GetNewPlaylistFunctions(), \&Plugins::PlaylistMan::SetNewPlaylistMode);
	Slim::Hardware::IR::addModeDefaultMapping('playlist', \%mapping);

	our $functref = Slim::Buttons::Playlist::getFunctions();
	$functref->{'addmenu'} = $functions{'addmenu'};
}

1;
