# Playlist Manager.pm by Philip Meyer (phil@hergest.demon.co.uk)
# Version 3.0
# This code is derived from code with the following copyright message:
#
# SlimServer Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2.

use strict;

package Plugins::PlaylistMan::Plugin;

use Slim::Utils::Strings qw (string);
use Slim::Utils::Misc;
use Slim::Utils::Log;
use Slim::Utils::Prefs;
use Slim::Player::Client;
use Slim::Player::Playlist;
use File::Spec::Functions qw(:ALL);

# Export the version to the server
use vars qw($VERSION);
$VERSION = "3.1";

use constant PLAYLISTMAN_NOW_PLAYING_MENU => 'PlaylistMan.NowPlayingAddMenu';
use constant PLAYLISTMAN_MENU_ADD => 'PlaylistMan.SelectPlaylistMenu';
use constant PLAYLISTMAN_SELECT_OTHER_PLAYLIST => 'PlaylistMan.SelectOtherPlaylist';

my @FavouritePlaylists = ('Bad Music', 'Bad Tags', 'Morning Alarm');

# A logger we will use to write plugin-specific messages. 
my $log = Slim::Utils::Log->addLogCategory( 
 { 
     'category'     => 'plugin.playlistman', 
     'defaultLevel' => 'WARN', 
     'description'  => 'PLUGIN_PLAYLISTMAN_NAME' 
 }
);


our @LegalChars = (
	undef, # placeholder for rightarrrow
	'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
	'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
	'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
	' ',
	'.', '-', '_',
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9'
);

our @legalMixed = (
	[' ','0'], 				                  # 0
	['.','-','_','1'], 			            # 1
	['a','b','c','A','B','C','2'],		   # 2
	['d','e','f','D','E','F','3'], 		   # 3
	['g','h','i','G','H','I','4'], 		   # 4
	['j','k','l','J','K','L','5'], 		   # 5
	['m','n','o','M','N','O','6'], 		   # 6
	['p','q','r','s','P','Q','R','S','7'], # 7
	['t','u','v','T','U','V','8'], 		   # 8
	['w','x','y','z','W','X','Y','Z','9'] 	# 9
);

our %context = ();

my %positions = ();
my $playlistindex;
my $song;

sub getDisplayName()
{
	return substr ($::VERSION, 0, 1) >= 6 ? 'PLUGIN_PLAYLISTMAN_NAME' : string('PLUGIN_PLAYLISTMAN_NAME');
}

sub enabled
{
	return 1;
}

sub setMode
{
   my $class = shift;
	my $client = shift;
	my $method = shift || '';

	# Handle request to exit our mode.
	if ($method eq 'pop')
	{
		# Pop the current mode off the mode stack and restore the previous one
		Slim::Buttons::Common::popMode($client);
		return;
	}

	$positions {$client} = 0 if (!defined $positions{$client});

	$client->lines(\&lines);
}

sub lines
{
	my $client = shift;
	my ($line1, $line2);

	my @addoptions = ("Display Bad Tags", "Display Bad Music");

	$line1 = "Display Playlists";
	$line2 = $addoptions[$positions{$client}];

	return ($line1, $line2);
}

our %functions =
(
	'left' => sub  {
		my $client = shift;
		Slim::Buttons::Common::popModeRight($client);
	},
	'right' => sub {
		my $client = shift;
		$client->bumpRight();
	},
	'up' => sub {
		my $client = shift;
		my $newPos = Slim::Buttons::Common::scroll($client, -1, 2, $positions{$client});
		$positions{$client} = $newPos;
		$client->update();
	},
	'down' => sub {
		my $client = shift;
		my $newPos = Slim::Buttons::Common::scroll($client, +1, 2, $positions{$client});
		$positions{$client} = $newPos;
		$client->update();
	},
	'addmenu' => sub
	{
		my $client = shift;
		$playlistindex = Slim::Buttons::Playlist::browseplaylistindex($client);
		$song = Slim::Player::Playlist::song($client, $playlistindex);
		my $songName = $song->title;

		$log->info("PlaylistMan: display addmenu for current playlist track number $playlistindex - $songName\n");

		my %params =
		(
			# The header (first line) to display whilst in this mode.
			header => '{PLUGIN_PLAYLISTMAN_NAME} {count}',

			# A reference to the list of items to display.
			listRef => [
            $client->string('PLUGIN_PLAYLISTMAN_ADD_TRACK_TO'),
            $client->string('PLUGIN_PLAYLISTMAN_ZAP_TRACK'),
            $client->string('PLUGIN_PLAYLISTMAN_SAVE_NEW_PLAYLIST')
            ],

			# A unique name for this mode that won't actually get displayed anywhere.
			modeName => PLAYLISTMAN_NOW_PLAYING_MENU,

			parentMode => Slim::Buttons::Common::mode($client),

			onPlay => sub {
				my ($client, $item) = @_;
				NowPlayingAddAction($client, $item);
			},

			# An anonymous function that is called every time the user presses the RIGHT button.
			onRight => sub {
				my ($client, $item) = @_;
				NowPlayingAddAction($client, $item);
			},

			# These are all menu items and so have a right-arrow overlay
			overlayRef => sub {
            my $client = shift; 
            return [ undef, $client->symbols('rightarrow') ];
         }
		);

		Slim::Buttons::Common::pushModeLeft($client, 'INPUT.Choice', \%params);
	}
);

sub NowPlayingAddAction
{
	my ($client, $item) = @_;

   $log->info("Now Playing Add Action: $item");

	if (Slim::Player::Playlist::count($client) > 0)
	{
		if ($item eq $client->string('PLUGIN_PLAYLISTMAN_ADD_TRACK_TO'))
		{
			my %params =
			(
				# The header (first line) to display whilst in this mode.
				header => '{PLUGIN_PLAYLISTMAN_ADD_TRACK_TO_PLAYLIST} {count}',

				# A reference to the list of items to display.
				listRef => [@FavouritePlaylists,
                        $client->string('PLUGIN_PLAYLISTMAN_OTHER_PLAYLIST'),
                        $client->string('PLUGIN_PLAYLISTMAN_NEW_PLAYLIST')],

				# A unique name for this mode that won't actually get displayed anywhere.
				modeName => PLAYLISTMAN_MENU_ADD,

				parentMode => Slim::Buttons::Common::mode($client),

				onPlay => sub {
					my ($client, $item) = @_;
					AddTrackToAction($client, $item);
				},

				onRight => sub {
					my ($client, $item) = @_;
					AddTrackToAction($client, $item);
				},

				# These are all menu items and so have a right-arrow overlay
				overlayRef => sub {
               my $client = shift; 
               return [ undef, $client->symbols('rightarrow') ];
            }
			);

			Slim::Buttons::Common::pushModeLeft($client, 'INPUT.Choice', \%params);
		}
		elsif ($item eq $client->string('PLUGIN_PLAYLISTMAN_ZAP_TRACK'))
		{
			$client->showBriefly(
            { 'line1' => $client->string('ZAPPING_FROM_PLAYLIST'),
              'line2' => Slim::Music::Info::standardTitle($client, $song)
            }
         );

			$client->execute(["playlist", "zap", $playlistindex]);
	 		Slim::Buttons::Common::popMode($client);
		}
		elsif ($item eq $client->string('PLUGIN_PLAYLISTMAN_SAVE_NEW_PLAYLIST'))
		{
   		my $serverPrefs = preferences('server');
			if (!$serverPrefs->get('playlistdir'))
			{
			   my $errormsg = $client->string('NO_PLAYLIST_DIR');
            $log->error($errormsg);
   			$client->showBriefly(
               { 'line1' => $client->string('PLUGIN_PLAYLISTMAN_NAME'),
                 'line2' => $errormsg
               }
            );
         }
			else
			{
				$context{$client} = $client->currentPlaylist ? Slim::Music::Info::standardTitle($client, $client->currentPlaylist) : 'A';
				Slim::Buttons::Common::pushModeLeft($client, 'SavePlaylist');
			}
		}
	}
}

sub AddTrackToAction
{
	my ($client, $item) = @_;

	if ($item eq $client->string('PLUGIN_PLAYLISTMAN_OTHER_PLAYLIST'))
	{
		SelectPlaylistToAddTo($client);
	}
	elsif ($item eq $client->string('PLUGIN_PLAYLISTMAN_NEW_PLAYLIST'))
	{
		Slim::Buttons::Common::pushModeLeft($client, 'NewPlaylist');
	}
	else
	{
		AddTrackTo($client, $item);
	}
}

sub SelectPlaylistToAddTo
{
	my $client = shift;

   $log->info("Select playlist to add track to");
   
	my @dirItems=();

	my $playlists = Slim::Schema->rs('Playlist')->browse;
	my $playlistsCount = 0;
	while (my $playlistObj = $playlists->next)
	{
		$log->info("PlaylistMan: Found playlistname=$playlistObj\n");

		#TODO: Remove configured common playlists from the list of selectable playlists?
		push @dirItems, $playlistObj->title;

		$playlistsCount = $playlistsCount+1;
	}
	
	$log->info("Number of playlists: $playlistsCount\n");
		
	if ($playlistsCount > 0)
	{
		my %params =
		(
			# The header (first line) to display whilst in this mode.
			# Handle large Font setting
			header => '{PLUGIN_PLAYLISTMAN_ADD_TRACK_TO_OTHER_PLAYLIST} {count}',
			listRef => \@dirItems,
			stringHeader => 1,

			# A unique name for this mode that won't actually get displayed anywhere.
			modeName => PLAYLISTMAN_SELECT_OTHER_PLAYLIST,
			parentMode => Slim::Buttons::Common::mode($client),
			onPlay => sub {
					my ($client, $item) = @_;
					AddTrackTo($client, $item);
				},

			# An anonymous function that is called every time the user presses the RIGHT button.
			onRight => sub {
					my ($client, $item) = @_;
					AddTrackTo($client, $item);
				},

			# These are all menu items and so have a right-arrow overlay
			overlayRef => 
				sub {
               my $client = shift;
               return [ undef, $client->symbols('rightarrow') ];
            }
      );

		Slim::Buttons::Common::pushModeLeft($client, 'INPUT.Choice', \%params);
	}
	else
	{
		$client->showBriefly({ 'line1' => $client->string('PLUGIN_PLAYLISTMAN_NO_PLAYLISTS')});
	}
}

sub AddTrackTo
{
	my ($client, $item) = @_;

	$log->info("Add Track To action: $item");
	
	if (Slim::Player::Playlist::count($client) > 0)
	{
		$client->showBriefly({'line1' => $client->string('PLUGIN_PLAYLISTMAN_ADDING_TO_PLAYLIST', $item), 'line2' => Slim::Music::Info::standardTitle($client, $song)});
		AddToPlaylist($client, $item, $song);

		Slim::Buttons::Common::popMode($client);
		Slim::Buttons::Common::popMode($client);
	}
}

sub AddToPlaylist
{
	my ($client, $playlist, $song) = @_;

	$log->info("Playlist name is: $playlist");

	my $playlistTitle = Slim::Music::Info::plainTitle($playlist, undef);

	$log->info("Playlist plaintitle is: '$playlistTitle'");

   my $serverPrefs = preferences('server');
	my $fileurl = Slim::Utils::Misc::fileURLFromPath(catfile($serverPrefs->get('playlistdir') . '/' . $playlist . '.m3u'));

	$log->info("fileurl=$fileurl");

	my $playlistObj = Slim::Schema->rs('Playlist')->updateOrCreate({
		'url' => $fileurl,
		'playlist' => $playlist,
		'attributes' => {
			'TITLE' => $playlistTitle,
			'CT'    => 'ssp',
		},
	});

	$playlistObj->appendTracks([ $song ]);

	Slim::Player::Playlist::scheduleWriteOfPlaylist($client, $playlistObj);
}

sub getFunctions
{
	return \%functions;
}


#Add to new playlist functionality

sub SetNewPlaylistMode
{
	my ($client, $push) = @_;

	$log->info("Display New Playlist menu");

	$client->lines(\&NewPlaylistLines);

   my $serverPrefs = preferences('server');
	if (!$serverPrefs->get('playlistdir'))
	{
	   my $errormsg = $client->string('NO_PLAYLIST_DIR');
      $log->error($errormsg);
		$client->showBriefly(
         { 'line1' => $client->string('PLUGIN_PLAYLISTMAN_NAME'),
           'line2' => $errormsg
         }
      );
	}
	elsif ($push ne 'push')
	{
		#my $playlist = '';
	}
	elsif ($client->modeParam('playlist') ne '')
   {
   	# don't do anything if we have a playlist name, since this
		# means we've done the text entry
   }
   else
	{
		# default to the existing title for a known playlist, otherwise just start with 'A'
		$context{$client} = $client->currentPlaylist ? Slim::Music::Info::standardTitle($client, $client->currentPlaylist) : 'A';

		Slim::Buttons::Common::pushMode($client,'INPUT.Text', {
			'callback' => \&Plugins::PlaylistMan::Plugin::NewPlaylistCallback,
			'valueRef' => \$context{$client},
			'charsRef' => \@LegalChars,
			'numberLetterRef' => \@legalMixed,
			'header' => $client->string('PLUGIN_PLAYLISTMAN_NEW_PLAYLIST'),
			'cursorPos' => 0
		});
	}
}

sub NewPlaylistLines
{
	my $client = shift;

	my ($line1, $line2, $arrow);

	$line1 = $client->string('PLUGIN_PLAYLISTMAN_ADD_TO_NEW_PLAYLIST');
	$line2 = $context{$client};
	$arrow = $client->symbols('rightarrow');

	return
	{
		'line1'   => $line1,
		'line2'   => $line2, 
		'overlay' => [ undef, $arrow ]
	};
}

our %NewPlaylistFunctions = (
	'left' => sub  {
		my $client = shift;
      $log->info("NewPlaylistFunctions - left");
		Slim::Buttons::Common::popModeRight($client);
	},
	'right' => sub  {
		my $client = shift;
		my $playlistfile = $context{$client};

      $log->info("NewPlaylistFunctions - right");

		Slim::Buttons::Common::setMode($client, 'playlist');
		NewPlaylist($client, $playlistfile);
	},
	'save' => sub {
		my $client = shift;
      $log->info("NewPlaylistFunctions - save");
		Slim::Buttons::Common::pushModeLeft($client, 'NewPlaylist');
	},
);

sub GetNewPlaylistFunctions
{
	return \%NewPlaylistFunctions;
}

sub NewPlaylist
{
	my ($client, $playlist) = @_;

	$log->info("Add to new playlist $playlist");

	AddToPlaylist($client, $playlist, $song);

	$client->showBriefly( {
		'line1' => $client->string('PLAYLIST_SAVING'),
		'line2' => $playlist
	});
}

sub NewPlaylistCallback
{
	my ($client, $type) = @_;

   $log->info("NewPlaylistCallback $type");

	if ($type eq 'nextChar')
   {
		# re-enter plugin with the new playlist title to get the confirmation screen for saving the playlist.
		Slim::Buttons::Common::pushModeLeft($client, 'NewPlaylist', { 'playlist' => $context{$client} });
	}
   elsif ($type eq 'backspace')
   {
   	Slim::Buttons::Common::popModeRight($client);
		Slim::Buttons::Common::popModeRight($client);
	}
   else
   {
		$client->bumpRight();
	}
}


#Save Playlist functionality

sub SetSavePlaylistMode
{
	my ($client, $push) = @_;

	$client->lines(\&SavePlaylistLines);

   my $serverPrefs = preferences('server');
	if (!$serverPrefs->get('playlistdir'))
	{
	   my $errormsg = $client->string('NO_PLAYLIST_DIR');
      $log->error($errormsg);
		$client->showBriefly(
         { 'line1' => $client->string('PLUGIN_PLAYLISTMAN_NAME'),
           'line2' => $errormsg
         }
      );
	}
	elsif ($push ne 'push')
	{
		# my $playlist = '';
	}
	elsif ($client->modeParam('playlist') ne '')
   {
   	# don't do anything if we have a playlist name, since this
		# means we've done the text entry
   }
	else
	{
		$context{$client} = $client->currentPlaylist ? Slim::Music::Info::standardTitle($client, $client->currentPlaylist) : 'A';

		# set cursor position to end of playlist title if the playlist is known
		my $cursorpos = $client->currentPlaylist ?  length($context{$client}) : 0;

		Slim::Buttons::Common::pushMode($client,'INPUT.Text', {
			'callback' => \&Plugins::PlaylistMan::Plugin::SavePlaylistCallback,
			'valueRef' => \$context{$client},
			'charsRef' => \@LegalChars,
			'numberLetterRef' => \@legalMixed,
			'header' => $client->string('PLAYLIST_AS'),
			'cursorPos' => $cursorpos
		});
	}
}

sub SavePlaylistLines
{
	my $client = shift;

	my ($line1, $line2, $arrow);

   my $serverPrefs = preferences('server');
   my $playlistdir = $serverPrefs->get('playlistdir');
	if (!$playlistdir)
	{
	   my $errormsg = $client->string('NO_PLAYLIST_DIR');
      $log->error($errormsg);

		$line1 = $client->string('NO_PLAYLIST_DIR');
		$line2 = $client->string('NO_PLAYLIST_DIR_MORE');
	}
   else
   {
   	my $newUrl = Slim::Utils::Misc::fileURLFromPath(catfile($playlistdir) . '/' . $context{$client} . '.m3u');
      $log->info("New playlist URL is $newUrl");

      if (Slim::Schema->rs('Track')->objectForUrl($newUrl))
      {
   		# Special text for overwriting an existing playlist
   		# if large text, make sure we show the message instead of the playlist name
   		if ($client->linesPerScreen == 1) {
   			$line2 = $client->doubleString('PLAYLIST_OVERWRITE');
   		} else {
   			$line1 = $client->string('PLAYLIST_OVERWRITE');
   			$line2 = $context{$client};
   		}
   		
   		$arrow = $client->symbols('rightarrow');
   	}
   	else
   	{
   		$line1 = $client->string('PLUGIN_PLAYLISTMAN_SAVE_NEW_PLAYLIST_CONFIRM');
   		$line2 = $context{$client};
   		$arrow = $client->symbols('rightarrow');
   	}
   }

	return
	{
		'line1'   => $line1,
		'line2'   => $line2, 
		'overlay' => [ undef, $arrow ]
	};
}

our %SavePlaylistFunctions = (
	'left' => sub  {
		my $client = shift;
		Slim::Buttons::Common::popModeRight($client);
	},
	'right' => sub  {
		my $client = shift;
		my $playlistfile = $context{$client};
		Slim::Buttons::Common::setMode($client, 'playlist');
		SavePlaylist($client, $playlistfile);
	},
	'save' => sub {
		my $client = shift;
		Slim::Buttons::Common::pushModeLeft($client, 'SavePlaylist');
	},
);

sub GetSavePlaylistFunctions
{
	return \%SavePlaylistFunctions;
}

sub SavePlaylist
{
	my $client = shift;
	my $playlistfile = shift;

	$client->execute(['playlist', 'save', $playlistfile]);
	$client->showBriefly( {
		'line1' => $client->string('PLAYLIST_SAVING'),
		'line2' => $playlistfile
	});
}

sub SavePlaylistCallback
{
	my ($client, $type) = @_;

	if ($type eq 'nextChar')
   {
		# re-enter plugin with the new playlist title to get the confirmation screen for saving the playlist.
		Slim::Buttons::Common::pushModeLeft($client, 'SavePlaylist', { 'playlist' => $context{$client}, });
	}
   elsif ($type eq 'backspace')
   {
   	Slim::Buttons::Common::popModeRight($client);
		Slim::Buttons::Common::popModeRight($client);
	}
   else
   {
		$client->bumpRight();
	}
}


####################################################################
# Adds a mapping for 'addmenu' function in Now Playing mode.
####################################################################
our %mapping = ('add.hold' => 'addmenu');

sub defaultMap
{
	return \%mapping; 
}

sub initPlugin
{
   $log->info("Initialse");

	Slim::Buttons::Common::addMode('SavePlaylist', GetSavePlaylistFunctions(), \&Plugins::PlaylistMan::Plugin::SetSavePlaylistMode);
	Slim::Buttons::Common::addMode('NewPlaylist', GetNewPlaylistFunctions(), \&Plugins::PlaylistMan::Plugin::SetNewPlaylistMode);
	Slim::Hardware::IR::addModeDefaultMapping('playlist', \%mapping);

	our $functref = Slim::Buttons::Playlist::getFunctions();
	$functref->{'addmenu'} = $functions{'addmenu'};
}
