#line 1 "Plugins/SongFileViewer/Plugin.pm"
# SongFileViewer::Plugin.pm by Philip Meyer
#
# This code is derived from code with the following copyright message:
#
# SliMP3 Server Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2.
#
# Changelog
# 1.0 - initial release

use strict;

package Plugins::SongFileViewer::Plugin;
use base qw(Slim::Plugin::Base);

use vars qw($VERSION);

use File::Basename;
use Slim::Utils::Strings qw (string);
use Slim::Utils::Prefs;

$VERSION = substr(q$Revision: 1.5$, 10);

use constant SONGFILEVIEWER_MENU => 'SongFileViewer.SongFileViewerMenu';

# A logger we will use to write plugin-specific messages. 
my $log = Slim::Utils::Log->addLogCategory( 
 { 
     'category'     => 'plugin.SongFileViewer', 
     'defaultLevel' => 'WARN',
     'description'  => 'PLUGIN_SONGFILEVIEWER' 
 }
);

sub getDisplayName { return 'PLUGIN_SONGFILEVIEWER'; }

sub initPlugin {
	my $class = shift;
	$class->SUPER::initPlugin(shift);
}

sub setMode {
	my $class = shift;
	my $client = shift;
	my $method = shift;

	if ($method eq 'pop') {
		Slim::Buttons::Common::popMode($client);
		return;
	}

	DisplayFileList($client);
}

my %songfileitems;

sub DisplayFileList
{
	my $client = shift;
	
	my $playlistindex = Slim::Buttons::Playlist::browseplaylistindex($client);
	my $song = Slim::Player::Playlist::song($client, $playlistindex);
	my $folder = dirname(Slim::Utils::Misc::pathFromFileURL($song->url));

	$log->debug("Song folder is $folder");
	
	# Extract folder name from the URL if it is a file.
	# find all text files within the folder. 
	FindSongFiles($folder);
	
	my @names = sort keys %songfileitems;

	my %params =
	(
		# The header (first line) to display whilst in this mode.
		header => '{PLUGIN_SONGFILEVIEWER} {count}',

		# A reference to the list of items to display.
		listRef => \@names,

		# A unique name for this mode that won't actually get displayed anywhere.
		modeName => SONGFILEVIEWER_MENU,

		parentMode => Slim::Buttons::Common::mode($client),

		onPlay => sub {
			my ($client, $name) = @_;
			DisplayFile($client, $folder, $name);
		},

		# An anonymous function that is called every time the user presses the RIGHT button.
		onRight => sub {
			my ($client, $item) = @_;
			DisplayFile($client, $folder, $item);
		},

		# These are all menu items and so have a right-arrow overlay
		overlayRef => sub {
         my $client = shift; 
         return [ undef, $client->symbols('rightarrow') ];
      }
	);

	Slim::Buttons::Common::pushModeLeft($client, 'INPUT.Choice', \%params);
}

sub FindSongFiles
{
	my $rootdir = shift;
	my $subdir = shift;

	my $dir = $rootdir;
	if (defined $subdir)
	{
		$dir = "$dir/$subdir";
	}

	opendir DIR, $dir or $log->error("Couldn't open song folder $dir");
	my @files = readdir DIR;

	foreach my $item (@files)
	{
		next if ($item =~ /^\./);

		my $p = "$dir/$item";

		if (-d $p)
		{
			if (defined $subdir)
			{
				$subdir = $subdir . "/$item";
			}
			else
			{
				$subdir = $item;
			}
			 
			$log->debug("search subfolder $subdir in $rootdir");

			FindSongFiles($rootdir, $subdir);
		}
		elsif ($item =~ /(.*)(\.txt)$/)
		{
			my $name = $1;

			$log->debug("Found $name in $item");

			$songfileitems{$name} = (defined $subdir) ? "$subdir/$item" : "$item";
		}
	}
	closedir DIR;
}

sub DisplayFile
{
	my ($client, $folder, $name) = @_;

	$log->info("SongFileViewer: Display $name");

	my $lines = ReadFile($client, $folder . '/' . $songfileitems{$name});

	my %params = (
		stringHeader => 1,
		header => $name,
		headerAddCount => 1,
		listRef => $lines,
		parentMode => Slim::Buttons::Common::mode($client),		  
	);

	Slim::Buttons::Common::pushModeLeft($client, 'INPUT.List', \%params);
}

my %SongFileViewerFunctions = (
	'done' => sub {
		my ( $client, $funct, $functarg ) = @_;
		Slim::Buttons::Common::popMode($client);
		$client->update();

		# pass along ir code to new mode if requested
		if ( defined $functarg && $functarg eq 'passback' ) {
			Slim::Hardware::IR::resendButton($client);
		}
	}
);

sub getFunctions {
	return \%SongFileViewerFunctions;
}

sub ReadFile {
	my ($client, $filepath) = @_;
	
	my @lines;

	$log->info("SongFileViewer: Read File $filepath");

	if (open MYFILE, $filepath) {
		@lines = <MYFILE>;
		close MYFILE;
	}
	else {
		push @lines, $client->string('PLUGIN_SONGFILEVIEWER_NOFILE') . ": $filepath";
	}

	return \@lines;
}

1;
