#line 1 "Plugins/SongFileViewer/Plugin.pm"
# SongFileViewer::Plugin.pm by Philip Meyer
#
# This code is derived from code with the following copyright message:
#
# SliMP3 Server Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2.

# TODO:
# * Support diacritics in file paths.

use strict;

package Plugins::SongFileViewer::Plugin;
use base qw(Slim::Plugin::Base);

use vars qw($VERSION);

use File::Basename;
use Slim::Utils::Strings qw (string);
use Slim::Utils::Prefs;

use constant SONGFILEVIEWER_MENU => 'SongFileViewer.SongFileViewerMenu';

# A logger we will use to write plugin-specific messages. 
my $log = Slim::Utils::Log->addLogCategory( 
 { 
     'category'     => 'plugin.SongFileViewer', 
     'defaultLevel' => 'WARN',
     'description'  => 'PLUGIN_SONGFILEVIEWER' 
 }
);

sub getDisplayName { return 'PLUGIN_SONGFILEVIEWER'; }

sub initPlugin {
	my $class = shift;
	my $client = shift;

	$VERSION = $class->_pluginDataFor('version');

	$log->info("Initialising " . Slim::Utils::Strings::string('PLUGIN_SONGFILEVIEWER') . " version $VERSION");

	$class->SUPER::initPlugin(shift);

	# requires Client
	# is a Query
	# has Tags
	# Function to call
	Slim::Control::Request::addDispatch(['songfileviewer', 'list'], [1, 1, 1, \&cliListHandler]);
	Slim::Control::Request::addDispatch(['songfileviewer', 'view'], [1, 1, 1, \&cliViewHandler]);

	Slim::Buttons::Common::addMode(SONGFILEVIEWER_MENU, {}, \&Plugins::SongFileViewer::Plugin::SongFileViewerMenu);

	if ($::VERSION ge '7.4') {
		Slim::Menu::AlbumInfo->registerInfoProvider( songFileViewer => (
			func => \&albumInfoHandler,
		) );
	}
	
	if ($::VERSION ge '7.1') {
		Slim::Menu::TrackInfo->registerInfoProvider( songFileViewer => (
			before => 'top',
			func => \&trackInfoHandler,
		) );
	}
}

sub setMode {
	my $class = shift;
	my $client = shift;
	my $method = shift;

	if ($method eq 'pop') {
		Slim::Buttons::Common::popMode($client);
		return;
	}

	my $playlistindex = Slim::Buttons::Playlist::browseplaylistindex($client);
	my $song = Slim::Player::Playlist::song($client, $playlistindex);

	my $folder = dirname(Slim::Utils::Misc::pathFromFileURL($song->url));

	DisplayFileList($client, $folder);
}

my %fileitems;

sub DisplayFileList
{
	my $client = shift;
	my $folder = shift;

	my @names = GetFileList($folder);
	
	my %params =
	(
		# The header (first line) to display whilst in this mode.
		header => '{PLUGIN_SONGFILEVIEWER} {count}',

		# A reference to the list of items to display.
		listRef => \@names,

		# A unique name for this mode that won't actually get displayed anywhere.
		modeName => SONGFILEVIEWER_MENU,

		parentMode => Slim::Buttons::Common::mode($client),

		onPlay => sub {
			my ($client, $name) = @_;
			DisplayFile($client, $folder, $name);
		},

		# An anonymous function that is called every time the user presses the RIGHT button.
		onRight => sub {
			my ($client, $name) = @_;
			DisplayFile($client, $folder, $name);
		},

		# These are all menu items and so have a right-arrow overlay
		overlayRef => sub {
         my $client = shift; 
         return [ undef, $client->symbols('rightarrow') ];
      }
	);

	Slim::Buttons::Common::pushModeLeft($client, 'INPUT.Choice', \%params);
}

sub GetFileList {
	my $folder = shift;

	$log->debug("Folder is $folder");

	# Extract folder name from the URL if it is a file.
	# find all text files within the folder. 
	%fileitems = ();
	FindFiles($folder);

	my @names = sort keys %fileitems;

	return @names;
}

sub FindFiles
{
	my $rootdir = shift;
	my $subdir = shift;

	my $dir = (defined $subdir) ? "$rootdir/$subdir" : "$rootdir";

	opendir DIR, $dir or $log->error("Couldn't open folder $dir");
	my @files = readdir DIR;

	foreach my $item (@files)
	{
		next if ($item =~ /^\./);

		my $p = "$dir/$item";

		if (-d $p)
		{
			my $searchdir = (defined $subdir) ? "$subdir/$item" : "$item";
			 
			$log->debug("search subfolder $searchdir in $rootdir");

			FindFiles($rootdir, $searchdir);
		}
		elsif ($item =~ /(.*)(\.txt)$/)
		{
			my $name = $1;

			$log->debug("Found $name in $item");

			$fileitems{$name} = (defined $subdir) ? "$subdir/$item" : "$item";
		}
	}
	closedir DIR;
}

sub DisplayFile
{
	my ($client, $folder, $name) = @_;

	$log->info("SongFileViewer: Display $name");

	my $file = $folder . '/' . $fileitems{$name};
 	my @lines = ReadFile($client, $file);

	my $item = {
		title        => $name,
		description  => join("\n\r", @lines)
	};

	Slim::Buttons::XMLBrowser::displayItemDescription($client, $item);
}

sub ReadFile {
	my ($client, $filepath) = @_;
	
	my @lines;

	$log->info("SongFileViewer: Read File $filepath");

	if (open MYFILE, $filepath) {
		@lines = <MYFILE>;
		close MYFILE;
	}
	else {
		push @lines, $client->string('PLUGIN_SONGFILEVIEWER_NOFILE') . ": $filepath";
	}

	return @lines;
}


my %SongFileViewerFunctions = (
	'done' => sub {
		my ( $client, $funct, $functarg ) = @_;
		Slim::Buttons::Common::popMode($client);
		$client->update();

		# pass along ir code to new mode if requested
		if ( defined $functarg && $functarg eq 'passback' ) {
			Slim::Hardware::IR::resendButton($client);
		}
	}
);

sub getFunctions {
	return \%SongFileViewerFunctions;
}

sub albumInfoHandler {
	my $return = objectInfoHandler( @_, 'album_id' );
	return $return;
}

sub trackInfoHandler {
	my $return = objectInfoHandler(@_, 'track_id');
	return $return;
}

sub objectInfoHandler {
	my ( $client, $url, $obj, $remoteMeta, $tags, $field ) = @_;

	return unless $client;

	$log->debug("objectInfoHandler for $field");

	my $id = $obj ? $obj->id : $tags->{search};

#Only works when $obj is a track...
	my $folder = dirname(Slim::Utils::Misc::pathFromFileURL($obj->url));
	my @names = GetFileList($folder);

	if (@names > 0) {
		return {
			type => 'redirect',
			name => $client->string('PLUGIN_SONGFILEVIEWER'),
			favorites => 0,
	
			player => {
				mode => SONGFILEVIEWER_MENU,
				modeParams => {
					$field => $id,
					obj => $obj,
				},
			},

			web  => {
				url => "plugins/SongFileViewer/index.html?$field=$id",
			},

	  		jive => {
	  			actions => {
	  				go => {
	  					cmd => [ 'songfileviewer', 'list' ],
	  					params => {
	  						menu => 1,
							$field => $id,
	  					}					
	  				}
	  			},
			},
		};
	}
}

sub SongFileViewerMenu
{
	my $client = shift;
	my $method = shift;

	$log->debug("SongFileViewerMenu: method=$method");

	if ($method eq 'pop') {
		Slim::Buttons::Common::popMode($client);
		return;
	}

	my $url;

	if ($client->modeParam('album_id')) {
		# find first track on the album
	}

	if ($client->modeParam('track_id')) {
		my $trackId = $client->modeParam('track_id');
		my $track = Slim::Schema->find(Track => $trackId);
		$url = $track->url;
	}

	my $folder = dirname(Slim::Utils::Misc::pathFromFileURL($url));

	$log->debug("SongFileViewerMenu: folder=$folder");
	Plugins::SongFileViewer::Plugin::DisplayFileList($client, $folder);
}

sub cliListHandler {
	$log->debug("Entering cliListHandler");
	my $request = shift;

	if ($request->isNotQuery([['songfileviewer'], ['list']])) {
		$log->warn("Incorrect query: " . $request->getRequestString());
		$request->setStatusBadDispatch();
		return;
	}

	my $client = $request->client();

	if(!defined $client) {
		$log->warn("Client required");
		$request->setStatusNeedsClient();
		return;
	}

	my $menu = $request->getParam('menu');

	my $params = $request->getParamsCopy();
	for my $k (keys %$params) {
		$log->debug("Got: $k=".$params->{$k});
	}

	my $chunkCount = 0;

	my $track;
	my $trackId = $request->getParam('track_id');
	my $trackUrl = $request->getParam('track_url');
	if (defined $trackId) {
		$track = Slim::Schema->find(Track => $trackId);
	}
	elsif (defined $trackUrl) {
		$track = Slim::Schema->objectForUrl( { url => $trackUrl } );
	}
	else {
		# Get the currently playing song on the chosen client.
		#TODO: need to check if there is a current song

		my $playlistindex = Slim::Buttons::Playlist::browseplaylistindex($client);
		$track = Slim::Player::Playlist::song($client, $playlistindex);
	}

	if (defined $track) {
		# Display a list of files to be chosen for the specified song
		my $trackTitle = $track->title;
		$log->debug("Display list of files for $trackTitle");

		if ($menu) {
			$log->debug("Display menu of files for $trackTitle");

			my $folder = dirname(Slim::Utils::Misc::pathFromFileURL($track->url));
			my @names = GetFileList($folder);

			foreach my $filename (@names) {
				my $actionParams;
				$actionParams = { folder => $folder, name => $filename, };
	
				$request->addResultLoop('item_loop', $chunkCount, 'text', "$filename");

				my $actions = {
					add => {
						player => 0,
						cmd => [ 'songfileviewer', 'view' ],
						params => $actionParams,
					},
				};

				# play and go have same behavior as add
				$actions->{play} = $actions->{add};
				$actions->{go} = $actions->{add};
				$request->addResultLoop('item_loop', $chunkCount, 'actions', $actions);

				$chunkCount++;
			}

			$request->addResult('count', $chunkCount);
			$request->addResult('offset', 0);
		}
	}

	$request->setStatusDone();
	$log->debug("Exiting cliListHandler");
}

sub cliViewHandler {
	$log->debug("Entering cliViewHandler");
	my $request = shift;

	if ($request->isNotQuery([['songfileviewer', 'view']])) {
		$log->warn("Incorrect command: " . $request->getRequestString());
		$request->setStatusBadDispatch();
		return;
	}

	my $client = $request->client();

	if(!defined $client) {
		$log->warn("Client required");
		$request->setStatusNeedsClient();
		return;
	}

	$request->setStatusProcessing();

	my $params = $request->getParamsCopy();
	for my $k (keys %$params) {
		$log->debug("Got: $k=".$params->{$k});
	}

	my $folder = $request->getParam('folder');
	my $name = $request->getParam('name');

	if (defined $name) {
		# If file param present, display the content of the specified file
		my $file = $folder . '\\' . $name . '.txt';
		my @lines = ReadFile($client, $file);
	 	my $text = join("", @lines);
		$log->debug("Content of $file is:\n$text");

		$request->addResult('window', { textarea => $text, });
		$request->addResult('offset', 0);
 		$request->addResult('count', 0);
	}
	else {
		$log->warn("name param required");
		#TODO: Better error response?
		$request->setStatusBadDispatch();
		return;
	}

	$request->setStatusDone();
	$log->debug("Exiting cliViewHandler");
}

1;
